﻿using System;
using System.Threading.Tasks;
using Butterfly.Core;
using Butterfly.HabboHotel.Achievements;
using Butterfly.HabboHotel.Advertisements;
using Butterfly.HabboHotel.Catalogs;
using Butterfly.HabboHotel.GameClients;
using Butterfly.HabboHotel.Items;
using Butterfly.HabboHotel.Misc;
using Butterfly.HabboHotel.Navigators;
using Butterfly.HabboHotel.Roles;
using Butterfly.HabboHotel.RoomBots;
using Butterfly.HabboHotel.Rooms;
using Butterfly.HabboHotel.Support;
using Butterfly.Storage;
using Butterfly_Emulator.ServerManager;


namespace Butterfly.HabboHotel
{
    class Game
    {
        private GameClientManager ClientManager;
        private ModerationBanManager BanManager;
        private RoleManager RoleManager;
        private HelpTool HelpTool;
        private Catalog Catalog;
        private Navigator Navigator;
        private ItemManager ItemManager;
        private RoomManager RoomManager;
        private AdvertisementManager AdvertisementManager;
        private PixelManager PixelManager;
        private AchievementManager AchievementManager;
        private ModerationTool ModerationTool;
        private BotManager BotManager;
        private Task StatisticsThread;
        private Task ConsoleTitleTask;
        private NavigatorCache mNavicache;

        internal Game(int conns)
        {
            ClientManager = new GameClientManager(conns);

            if (ButterflyEnvironment.GetConfig().data["client.ping.enabled"] == "1")
            {
                ClientManager.StartConnectionChecker();
            }

            DateTime Start = DateTime.Now;
            Logging.WriteLine("Connecting to database...");
            using (DatabaseClient dbClient = ButterflyEnvironment.GetDatabase().GetClient())
            {
                TimeSpan TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Connected to database! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");
                //UberEnvironment.GameInstance = this;

                BanManager = new ModerationBanManager();
                RoleManager = new RoleManager();
                HelpTool = new HelpTool();
                Catalog = new Catalog();
                Navigator = new Navigator();
                ItemManager = new ItemManager();
                RoomManager = new RoomManager();
                AdvertisementManager = new AdvertisementManager();
                PixelManager = new PixelManager();
                AchievementManager = new AchievementManager();
                ModerationTool = new ModerationTool();
                BotManager = new BotManager();
            }
        }

        internal void ContinueLoading()
        {
            DateTime Start;
            TimeSpan TimeUsed;
            using (DatabaseClient dbClient = ButterflyEnvironment.GetDatabase().GetClient())
            {
                Start = DateTime.Now;
                BanManager.LoadBans(dbClient);
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Ban manager -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                //RoleManager.LoadRoles(dbClient);
                RoleManager.LoadRights(dbClient);
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Role manager -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                HelpTool.LoadCategories(dbClient);
                HelpTool.LoadTopics(dbClient);
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Help tool -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                Catalog.Initialize(dbClient);
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Catacache -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                Navigator.Initialize(dbClient);
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Navigator -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                ItemManager.LoadItems(dbClient);
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Item manager -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                RoomManager.LoadModels(dbClient);
                RoomManager.LoadDryCache();
                mNavicache = new NavigatorCache();
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Room manager -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                AdvertisementManager.LoadRoomAdvertisements(dbClient);
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Adviserment manager -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                PixelManager.Start();
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Pixel manager -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                AchievementManager.LoadAchievements(dbClient);
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Achievement manager -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                ModerationTool.LoadMessagePresets(dbClient);
                ModerationTool.LoadPendingTickets(dbClient);
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Moderation tool -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                BotManager.LoadBots(dbClient);
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Bot manager manager -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                Catalog.InitCache();
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Catalogue manager -> READY! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");

                Start = DateTime.Now;
                DatabaseCleanup(dbClient);
                TimeUsed = DateTime.Now - Start;
                Logging.WriteLine("Database -> Cleanup performed! (" + TimeUsed.Seconds + " s, " + TimeUsed.Milliseconds + " ms)");
            }
            StatisticsThread = new Task(LowPriorityWorker.Process);
            // StatisticsThread.Name = "Low Priority Worker";
            // StatisticsThread.Priority = ThreadPriority.Lowest;
            StatisticsThread.Start();
            ConsoleTitleTask = new Task(LowPriorityWorker.ConsoleTitleWorker);
            ConsoleTitleTask.Start();

            Logging.WriteLine("Game manager -> READY!");

            DataSocket.SetupListener(5485);
            DataSocket.Start();
        }

        internal static void DatabaseCleanup(DatabaseClient dbClient)  // WTf did u drunk or something it's always 1 (int serverStatus) LOL
        {
            ////dbClient.ExecuteQuery("UPDATE users SET online = '0'");
            ////dbClient.ExecuteQuery("UPDATE rooms SET users_now = '0'");
            //dbClient.ExecuteQuery("UPDATE user_roomvisits SET exit_timestamp = '" + UberEnvironment.GetUnixTimestamp() + "' WHERE exit_timestamp <= 0");
            dbClient.ExecuteQuery("UPDATE server_status SET status = '1', users_online = '0', rooms_loaded = '0', server_ver = '" + ButterflyEnvironment.PrettyVersion + "', stamp = '" + ButterflyEnvironment.GetUnixTimestamp() + "' LIMIT 1;");
            // todo: more db cleanup stuff?

            //Logging.WriteLine("Database -> Cleanup performed");
        }

        internal void Destroy()
        {
            //if (StatisticsThread != null)
            //{
            //    StatisticsThread = null;
            //}

            using (Butterfly.Storage.DatabaseClient dbClient = ButterflyEnvironment.GetDatabase().GetClient())
            {
                DatabaseCleanup(dbClient);
            }
            if (GetClientManager() != null)
            {
                //GetClientManager().Clear();
                GetClientManager().StopConnectionChecker();
            }

            if (GetPixelManager() != null)
            {
                PixelManager.KeepAlive = false;
            }

            //ClientManager = null;
            //BanManager = null;
            //RoleManager = null;
            //HelpTool = null;
            //Catalog = null;
            //Navigator = null;
            //ItemManager = null;
            //RoomManager = null;
            //AdvertisementManager = null;
            //PixelManager = null;

            Console.WriteLine("Destroyed Habbo Hotel.");
        }

        internal GameClientManager GetClientManager()
        {
            return ClientManager;
        }

        internal ModerationBanManager GetBanManager()
        {
            return BanManager;
        }

        internal RoleManager GetRoleManager()
        {
            return RoleManager;
        }

        internal HelpTool GetHelpTool()
        {
            return HelpTool;
        }

        internal Catalog GetCatalog()
        {
            return Catalog;
        }

        internal Navigator GetNavigator()
        {
            return Navigator;
        }

        internal ItemManager GetItemManager()
        {
            return ItemManager;
        }

        internal RoomManager GetRoomManager()
        {
            return RoomManager;
        }

        internal AdvertisementManager GetAdvertisementManager()
        {
            return AdvertisementManager;
        }

        internal PixelManager GetPixelManager()
        {
            return PixelManager;
        }

        internal AchievementManager GetAchievementManager()
        {
            return AchievementManager;
        }

        internal ModerationTool GetModerationTool()
        {
            return ModerationTool;
        }

        internal BotManager GetBotManager()
        {
            return BotManager;
        }

        internal NavigatorCache GetNaviCache()
        {
            return mNavicache;
        }
    }
}
